<?php
/**
 * Suppress "error - 0 - No summary was found for this file" on phpdoc generation
 *
 * @package WPDataAccess\Utilities
 */

namespace WPDataAccess\Utilities {

	/**
	 * Class WPDA_Import_File
	 *
	 * Loads the content of an import file and imports it.
	 *
	 * @package WPDataAccess\Utilities
	 * @author  Peter Schulz
	 * @since   1.0.0
	 */
	class WPDA_Import_File {

		/**
		 * Content of import file
		 *
		 * @var string
		 */
		protected $file_content;

		/**
		 * WPDA_Import constructor
		 *
		 * Save file content in member variable. Strip leading and trailing spaces.
		 *
		 * @since   1.0.0
		 *
		 * @param string $file_content Content of script (import) file.
		 */
		public function __construct( $file_content ) {

			$this->file_content = trim( $file_content );

		}

		/**
		 * Import file content
		 *
		 * Import method writes the content of the import file to the database. Security checks:
		 * + Only INSERT INTO is allowed: no other DML, DDL and DCL statements allowed
		 * + Only inserts into the table name provided are allowed
		 * + Use explain to check the number of tables affect: more than 1 looks like SQL injection
		 *
		 * Since wpdb-query() only processes one query at a time we only need to check the type of statement at the
		 * beginning of the script ($file_content).
		 *
		 * This method only return -1 if a failure occurs or the number of rows inserted. The number of probable error
		 * cause is to huge and complex to check all possibilities. Exports created from the WP Data Access table list
		 * should normally import without problems. For manually created imports responsibility is with the developer.
		 *
		 * @since   1.0.0
		 *
		 * @param string $schema_name Schema in which this import allowes inserts.
		 * @param string $table_name Table in which this import allowes inserts.
		 * @param string $hide_errors ON = hide errors, OFF = show errors.
		 * @return int Either return -1 (indicates failure) or the number of records inserted.
		 */
		public function import( $schema_name, $table_name, $hide_errors ) {

			// Write file content to array for security check (150 characters is sufficient to check DML and table name).
			$dml_check = explode( ' ', substr( $this->file_content, 0, 150 ) );

			if ( ! isset( $dml_check[0] ) || ! isset( $dml_check[1] ) ) {
				// No content.
				if ( ( 'off' === $hide_errors ) ) {
					$msg = new WPDA_Message_Box(
						[
							'message_text'           => __( 'Nothing to import', 'wp-data-access' ),
							'message_type'           => 'error',
							'message_is_dismissible' => false,
						]
					);
					$msg->box();
				}
				return -1;
			}

			// Check first two words (must be insert into, no other statements allowed).
			if ( strtolower( $dml_check[0] . $dml_check[1] ) !== 'insertinto' ) {
				// Only insert into is allowed.
				if ( ( 'off' === $hide_errors ) ) {
					$msg = new WPDA_Message_Box(
						[
							'message_text'           => __( 'Only INSERT INTO allowed', 'wp-data-access' ),
							'message_type'           => 'error',
							'message_is_dismissible' => false,
						]
					);
					$msg->box();
				}
				return -1;
			}

			// Check table name (using stristr should cover backticks and schema_names as well).
			if ( ! stristr( $dml_check[2], $table_name ) ) {
				// Insert into other tables is not allowed.
				if ( ( 'off' === $hide_errors ) ) {
					$msg = new WPDA_Message_Box(
						[
							'message_text'           => sprintf( __( 'Only INSERT INTO table %s allowed', 'wp-data-access' ), $table_name ),
							'message_type'           => 'error',
							'message_is_dismissible' => false,
						]
					);
					$msg->box();
				}
				return -1;
			}

			global $wpdb;

			$suppress = $wpdb->suppress_errors( 'on' === $hide_errors ); // Suppres errors.

			// Execute an explain statement to check the number of tables affected (max 1 accepted!).
			$check = $wpdb->get_results( "explain {$this->file_content}", 'ARRAY_A' ); // WPCS: unprepared SQL OK; db call ok; no-cache ok.
			if ( count( $check ) > 1 ) {
				// Multiple statements found.
				if ( ( 'off' === $hide_errors ) ) {
					$msg = new WPDA_Message_Box(
						[
							'message_text' => __( 'Only one SQL statement allowed', 'wp-data-access' ),
							'message_type' => 'error',
							'message_is_dismissible' => false,
						]
					);
					$msg->box();
				}
				return -1;
			}

			if ( '' === $schema_name ) {
				// Perform insert on WordPress schema.
				$result = $wpdb->query( $this->file_content ); // WPCS: unprepared SQL OK; db call ok; no-cache ok.
			} else {
				// Perform insert on other schema.
				$db = new \wpdb( DB_USER, DB_PASSWORD, $schema_name, DB_HOST );
				$result = $db->query( $this->file_content ); // WPCS: unprepared SQL OK; db call ok; no-cache ok.
				$db->close();
			}

			$wpdb->suppress_errors( $suppress );

			if ( ! $result ) {
				return -1; // Errors during script execution.
			} else {
				return $result; // Number of records added to table.
			}

		}

	}

}
