<?php

namespace WPDataAccess\Utilities {

	use WPDataAccess\WPDA;

	/**
	 * Class WPDA_Import_Multi
	 *
	 * Imports a script file that might contain multiple SQL statement including create table and insert into
	 * statements.
	 *
	 * @package WPDataAccess\Utilities
	 * @author  Peter Schulz
	 * @since   1.6.0
	 */
	class WPDA_Import_Multi {

	    const SOLUTIONS = '(<a href="https://wpdataaccess.com/import-file-too-large/">see solutions</a>)';

		/**
		 * URL where to post data
		 *
		 * @var string
		 */
		protected $url;

		/**
		 * Database schema name
		 *
		 * @var string
		 */
		protected $schema_name;

        /**
         * Content of import file
         *
         * @var string
         */
        protected $file_content;

		/**
		 * WPDA_Import constructor
		 *
		 * Checks if imports are allowed. Throws an exception if imports are not allowed.
		 *
		 * @since   1.6.0
		 *
		 * @param string $page Page where to post data (url).
		 * @param string $schema_name Database schema name.
		 * @throws \Exception Throws exception if export is disabled.
		 */
		public function __construct( $page, $schema_name ) {

			if ( 'on' !== WPDA::get_option( WPDA::OPTION_BE_ALLOW_IMPORTS ) ) {
				// Prevent import object being created: exception must be handled in calling method.
				throw new \Exception( __( 'ERROR: Not authorized', 'wp-data-access' ) );
			}

			$this->url         = $page;
			$this->schema_name = $schema_name;

		}

		/**
		 * Checks if request is valid and allowed
		 *
		 * If the requested import is valid and allowed, the import file is loaded and its content imported.
		 *
		 * @since   1.6.0
		 */
		public function check_post() {

			// Check if import was requested.
			if (
			        isset( $_REQUEST['action'] ) &&
                    'import' === sanitize_text_field( wp_unslash( $_REQUEST['action'] ) ) // input var okay.
			) {
				// Security check.
				$wp_nonce = isset( $_REQUEST['_wpnonce'] ) ? sanitize_text_field( wp_unslash( $_REQUEST['_wpnonce'] ) ) : '?'; // input var okay.
				if ( ! wp_verify_nonce( $wp_nonce, "wpda-import-from-data-explorer" ) ) {
					wp_die( __( 'ERROR: Not authorized', 'wp-data-access' ) );
				}

				if ( isset( $_FILES['filename'] ) ) {

					if ( 0 === $_FILES['filename']['error']
						&& is_uploaded_file( $_FILES['filename']['tmp_name'] )
					) {
					    if (
					    		'application/zip' === $_FILES['filename']['type'] ||
								'application/x-zip' === $_FILES['filename']['type'] ||
								'application/x-zip-compressed' === $_FILES['filename']['type']
						) {
					        // Process ZIP file.
                            if ( class_exists( '\ZipArchive' ) ) {
                                $zip = new \ZipArchive;
                                if ( $zip->open( $_FILES['filename']['tmp_name'] ) ) {
                                    for ( $i = 0; $i < $zip->numFiles; $i++ ) {
                                        $fp = $zip->getStream( $zip->getNameIndex( $i ) );
                                        $this->file_content = '';
                                        while( !feof( $fp ) ) {
                                            $this->file_content .= fread( $fp, 2 );
                                        }
                                        $this->import( $zip->getNameIndex( $i ) );
                                    }
                                } else {
                                    // Error reading ZIP file.
                                    $this->import_failed( sprintf( __( 'Import failed (error reading ZIP file)', 'wp-data-access' ), $_FILES['filename']['name'] ) );
                                }
                            } else {
                                // ZipArchive not installed.
                                $this->import_failed( sprintf( __( 'Import failed - ZipArchive not installed %s', 'wp-data-access' ), SELF::SOLUTIONS ) );
                            }
                        } else {
					        // Process plain file.
                            $this->file_content = file_get_contents( $_FILES['filename']['tmp_name'] );
                            $this->import( $_FILES['filename']['name'] );
                        }
					}
				} else {
                    $this->upload_failed();
				}
			} elseif ( isset( $_REQUEST['impchk'] ) ) {
                $this->upload_failed();
            }

		}

		protected function import( $object_name ) {

            // Check if errors should be shown.
            $hide_errors = isset( $_REQUEST['hide_errors'] ) ? $_REQUEST['hide_errors'] : 'off';

            $result = true;
            global $wpdb;

            if ( '' === $this->schema_name || $wpdb->dbname === $this->schema_name ) {
                $db = new \wpdb( DB_USER, DB_PASSWORD, $this->schema_name, DB_HOST );
				$db->suppress_errors( 'on' === $hide_errors );
            } else {
                $db = null;
				$suppress = $wpdb->suppress_errors( 'on' === $hide_errors );
            }

            $sqls = array_filter(explode(";\n", $this->file_content));
            foreach($sqls as $sql) {
                if ( null === $db ) {
                    if ( false === $wpdb->query( $sql ) ) {
                        $result = false;
                    }
                } else {
                    if ( false === $db->query( $sql ) ) {
                        $result = false;
                    }
                }
            }

            if ( null !== $db ) {
                $db->close();
            } else {
				$wpdb->suppress_errors( $suppress );
			}

			// Process file content.
            if ( ! $result ) {
                $this->import_failed( sprintf( __( 'Import %s failed (check import file)', 'wp-data-access' ), $object_name ) );
            } else {
                // Import succeeded.
                $msg = new WPDA_Message_Box(
                    [
                        'message_text' => sprintf( __( 'Import %s completed succesfully', 'wp-data-access' ), $object_name),
                    ]
                );
                $msg->box();
            }

        }

		protected function import_failed( $msg ) {

            // An error occured: inform user.
            $msg = new WPDA_Message_Box(
                [
                    'message_text'           => $msg,
                    'message_type'           => 'error',
                    'message_is_dismissible' => false,
                ]
            );
            $msg->box();

        }

		protected function upload_failed() {

            // File upload failed: inform user.
            $msg = new WPDA_Message_Box(
                [
                    'message_text'           => sprintf( __( 'File upload failed %s', 'wp-data-access' ), SELF::SOLUTIONS),
                    'message_type'           => 'error',
                    'message_is_dismissible' => false,
                ]
            );
            $msg->box();

        }

		/**
		 * Adds an import button
		 *
		 * @since   1.6.0
		 *
		 * @param string $label Button label.
		 * @param string $class Button CSS class.
		 */
		public function add_button( $label = '', $class = 'page-title-action' ) {

			if ( '' === $label ) {
				$label = __( 'Import', 'wp-data-access' );
			}

			?>

			<a href="javascript:void(0)"
			   onclick="jQuery('#upload_file_container_multi').show()"
			   class="<?php echo esc_attr( $class ); ?>">
				<?php echo esc_attr( $label ); ?></a>

			<?php

		}

		/**
		 * Adds an import container
		 *
		 * The container contains an upload form. The container is hidden by default. When the button created in
		 * {@see WPDA_Import_Multi::add_button()} is clicked, the container is shown.
		 *
		 * @since   1.6.0
		 */
		public function add_container() {

		    if ( class_exists( '\ZipArchive' ) ) {
		        $file_extentions = '.sql,.zip';
            } else {
                $file_extentions = '.sql';
            }

			?>

			<div id="upload_file_container_multi" style="display: none">
				<div>&nbsp;</div>
				<div class="wpda_upload">
					<p>
						<strong><?php echo __( 'INFO', 'wp-data-access' ); ?></strong>
					</p>
					<p class="wpda_list_indent">
						<?php
						$text = __(
							'This feature is meant to import files you have exported using the export feature of this plugin. You can import multiple tables at once. Import is performed without validity checks. Data is imported immediately after file upload.',
							'wp-data-access'
						);
						echo sprintf( esc_attr( $text ) );
						?>
					</p>
					<form method="post" action="<?php echo esc_attr( $this->url ); ?>&impchk"
						  enctype="multipart/form-data">
						<input type="file" name="filename" id="filename" accept="<?php echo esc_attr( $file_extentions ); ?>">
						<input type="submit" value="<?php echo __( 'Import file', 'wp-data-access' ); ?>"
							   class="button button-secondary"
							   onclick="if (jQuery('#filename').val() == '') { alert('<?php echo __( 'No file to import!', 'wp-data-access' ); ?>'); return false; } ">
						<a href="javascript:void(0)"
						   onclick="jQuery('#upload_file_container_multi').hide()"
						   class="button button-secondary"><?php echo __( 'Cancel', 'wp-data-access' ); ?></a>
						<label style="vertical-align:baseline;"><input type="checkbox" name="hide_errors" style="vertical-align:sub;" checked>Hide errors</label>
						<input type="hidden" name="action" value="import">
						<?php wp_nonce_field( "wpda-import-from-data-explorer", '_wpnonce', false ); ?>
					</form>
				</div>
				<div>&nbsp;</div>
			</div>

			<?php

		}

	}

}
