<?php
/**
 * Suppress "error - 0 - No summary was found for this file" on phpdoc generation
 *
 * @package WPDataAccess\Utilities
 */

namespace WPDataAccess\Utilities {

	use WPDataAccess\Design_Table\WPDA_Design_Table_Model;
	use WPDataAccess\User_Menu\WPDA_User_Menu_Model;
	use WPDataAccess\WPDA;

	/**
	 * Class WPDA_Repository
	 *
	 * Recreate repository objects.
	 *
	 * @package WPDataAccess\Utilities
	 * @author  Peter Schulz
	 * @since   1.0.0
	 */
	class WPDA_Repository {

		/**
		 * Repository version numbers
		 *
		 * Must match name of sub folder where scripts are located
		 */
		const VERSION_REPOSITORY_V1 = 'v1';
		const VERSION_REPOSITORY_V2 = 'v2';
		const VERSION_REPOSITORY_V3 = 'v3';

		/**
		 * Repository create script files
		 *
		 * Processed in the order in which they appear
		 */
		const CREATE_REPOSITORIES = [
			self::VERSION_REPOSITORY_V1 => [
				'create_table_menu_items.sql',
				'create_trigger_menu_items_before_insert.sql',
				'create_trigger_menu_items_before_update.sql',
			],
			self::VERSION_REPOSITORY_V2 => [
				'create_table_menu_items.sql',
				'create_trigger_menu_items_before_insert.sql',
				'create_trigger_menu_items_before_update.sql',
				'create_table_table_design.sql',
				'create_table_table_design_alter1.sql',
				'create_table_table_design_alter2.sql',
				'create_table_table_design_alter3.sql',
			],
			self::VERSION_REPOSITORY_V3 => [
				'backup_table_menu_items.sql',
				'drop_table_menu_items.sql',
				'create_table_menu_items.sql',
				'create_trigger_menu_items_before_insert.sql',
				'create_trigger_menu_items_before_update.sql',
				'restore_table_menu_items.sql',
				'cleanup_table_menu_items.sql',
				'create_table_table_design.sql',
				'create_table_table_design_alter1.sql',
				'create_table_table_design_alter2.sql',
				'create_table_table_design_alter3.sql',
			],
		];

		/**
		 * Repository drop script files
		 *
		 * Processed in the order in which they appear
		 */
		const DROP_REPOSITORIES = [
			self::VERSION_REPOSITORY_V1 => [
				'drop_table_menu_items.sql',
			],
			self::VERSION_REPOSITORY_V2 => [
				'drop_table_menu_items.sql',
				'drop_table_table_design.sql',
			],
			self::VERSION_REPOSITORY_V3 => [
				'drop_table_menu_items.sql',
				'drop_table_table_design.sql',
			],
		];

		/**
		 * Default version number
		 */
		const VERSION_CURRENT_REPOSITORY = self::VERSION_REPOSITORY_V3;

		/**
		 * Drop repository
		 *
		 * @since   1.0.0
		 *
		 * @param string $repository_version Repository version (use class constants).
		 */
		public function drop( $repository_version = self::VERSION_CURRENT_REPOSITORY ) {

			$this->run_scripts( self::DROP_REPOSITORIES[ $repository_version ], $repository_version );

		}

		/**
		 * Run script files
		 *
		 * Run scripts for a repository (create or drop). Scripts will be processed in the order in which they are
		 * added to the repository array.
		 *
		 * @since   1.0.0
		 *
		 * @param array  $sql_repository_files Repository array (use class constants).
		 * @param string $repository_version Repository version (use class constants).
		 */
		protected function run_scripts( $sql_repository_files, $repository_version ) {

			global $wpdb;

			$suppress = $wpdb->suppress_errors( true );

			$files_processed_successfully = [];
			$files_processed_with_errors  = [];

			foreach ( $sql_repository_files as $sql_repository_file_name ) {

				$sql_repository_file   = plugin_dir_path( dirname( __FILE__ ) ) .
					'../admin/repository/' . $repository_version . '/' . $sql_repository_file_name;
				$sql_repository_handle = fopen( $sql_repository_file, 'r' );

				if ( $sql_repository_handle ) {

					// Read file content and close handle.
					$sql_repository_file_content = fread( $sql_repository_handle, filesize( $sql_repository_file ) );
					fclose( $sql_repository_handle );

					// Replace WP prefix and WPDA prefix.
					$sql_repository_file_content = str_replace( '{wp_prefix}', $wpdb->prefix, $sql_repository_file_content );
					$sql_repository_file_content = str_replace( '{wpda_prefix}', WPDA::get_option( WPDA::OPTION_WPDA_PREFIX ), $sql_repository_file_content );

					// Run script.
					$result = $wpdb->query( $sql_repository_file_content ); // WPCS: unprepared SQL OK; db call ok; no-cache ok.

					if ( ! $result ) {
						$files_processed_with_errors[] = $sql_repository_file_name;
					} else {
						$files_processed_successfully[] = $sql_repository_file_name;
					}
				}
			}

			$wpdb->suppress_errors( $suppress );

		}

		/**
		 * Create repository
		 *
		 * @since   1.0.0
		 *
		 * @param string $repository_version Repository version (use class constants).
		 */
		public function create( $repository_version = self::VERSION_CURRENT_REPOSITORY ) {

			$this->run_scripts( self::CREATE_REPOSITORIES[ $repository_version ], $repository_version );

		}

		/**
		 * Inform user if repository is invalid
		 *
		 * Repository is valid if:
		 * + Table menu_items exists
		 * + Before insert trigger exists
		 * + Before update trigger exists
		 * + Table table_design exists
		 *
		 * Repository is invalid if:
		 * + Table menu_items does not exist
		 * + Table table_design does not exist
		 *
		 * Repository is incomplete but working:
		 * + Table menu_items exists
		 * + Triggers do not exists
		 *
		 * If the triggers do not exist but the table is available, inserts and updates are possible. The dbms however
		 * will not validate any data on insert and update. When users use WPDA_User_Menu_Form, data validation is
		 * perform on the server. If users use WPDA_Simple_Form data validation will be insufficient.
		 *
		 * @since   1.0.0
		 */
		public function inform_user() {

			if ( isset( $_REQUEST['setup_error'] ) && 'off' === $_REQUEST['setup_error'] ) {
				// Turn off menu management not available message.
				WPDA::set_option( WPDA::OPTION_WPDA_SETUP_ERROR, 'off' );
			}

			if ( 'off' !== WPDA::get_option( WPDA::OPTION_WPDA_SETUP_ERROR ) ) {
				if ( ! WPDA_User_Menu_Model::table_menu_items_exists() ||
				     ! WPDA_Design_Table_Model::table_table_design_exists() ) {
					$msg = new WPDA_Message_Box(
						[
							'message_text' =>
								__( 'WP Data Access menu management currently not available.', 'wp-data-access' ) .
								' ' .
								__( 'ACTION', 'wp-data-access' ) .
								': ' .
								__( 'Recreate', 'wp-data-access' ) .
								' ' .
								'<a href="?page=wpda_settings&tab=repository">' . __( 'repository', 'wp-data-access' ) . '</a>' .
								' ' .
								__( 'to to solve this problem.', 'wp-data-access' ) .
								' (' .
								'<a href="?' . $_SERVER['QUERY_STRING'] . '&setup_error=off">' . __( 'do not show this message again', 'wp-data-access' ) . '</a>' .
								')',
						]
					);

					$msg->box();
				}
			}

		}

	}

}
