<?php
/**
 * Suppress "error - 0 - No summary was found for this file" on phpdoc generation
 *
 * @package plugin\includes
 */

use WPDataAccess\Utilities\WPDA_Repository;
use WPDataAccess\Utilities\WPDA_Upgrade;
use WPDataAccess\WPDA;

/**
 * Class WP_Data_Access_Switch
 *
 * Switch to:
 * + activate plugin {@see WP_Data_Access_Switch::activate()}
 * + deactive plugin {@see WP_Data_Access_Switch::deactivate()}
 *
 * @package plugin\includes
 * @author  Peter Schulz
 * @since   1.0.0
 *
 * @see WP_Data_Access_Switch::activate()
 * @see WP_Data_Access_Switch::deactivate()
 */
class WP_Data_Access_Switch {

	/**
	 * Activate plugin WP Data Access
	 *
	 * The user must have the appropriate privileges to perform this operation.
	 *
	 * For single site installation {@see WP_Data_Access_Switch::activate_blog()} will be called. For multi site
	 * installations {@see WP_Data_Access_Switch::activate_blog()} must be called for every blog.
	 *
	 * IMPORTANT!!!
	 *
	 * For blogs installed on multi site installations after activation of the plugin, activation of the plugin for
	 * that blog will not be performed if the plugin is network activated. In that case the admin user of the blog
	 * will receive a message when viewing a plugin page with an option to follow these steps manually.
	 *
	 * @since   1.0.0
	 *
	 * @see WP_Data_Access_Switch::activate_blog()
	 */
	public static function activate() {

		if ( current_user_can( 'activate_plugins' ) ) {

			// Activate plugin.
			if ( is_multisite() ) {

				global $wpdb;

				// Multisite installation.
				$blogids = $wpdb->get_col( "select blog_id from $wpdb->blogs" ); // db call ok; no-cache ok.
				foreach ( $blogids as $blog_id ) {

					// Uninstall blog.
					switch_to_blog( $blog_id );
					self::activate_blog();
					restore_current_blog();

				}
			} else {

				// Single site installation.
				self::activate_blog();

			}
		} else {

			wp_die( esc_html__( 'ERROR: Not authorized', 'wp-data-access' ) );

		}

	}

	/**
	 * Activate blog
	 *
	 * The user must have the appropriate privileges to perform this operation.
	 *
	 * On activation this method checks whether there has previously been a version of the plugin installed. For this
	 * purpose the wp_options table read directly (usually done via class WPDA). If a value is found, this method
	 * checks if the version number in wp_options is the same as the plugin version. If these are equal, no action is
	 * needed. If they are not equal, this method will check if there is an upgrade or downgrade for the delta
	 * between these releases.
	 *
	 * This action is performed on the 'active WordPress blog'. On single site there is only one blog. On multisite
	 * installations it must be executed for every blog.
	 *
	 * On a fresh installation the following actions are performed:
	 * + save plugin version number in wp_options {@see WPDA::set_option()}
	 * + create plugin repository {@see WPDA_Repository::create()}
	 *
	 * The actions performed on an upgrade or downgrade need to be coded in class {@see WPDA_Upgrade}. The function
	 * name for the upgrade or download must reflect value $function_upgrade_name to be performed correctly. As there
	 * are no upgrades or downgrades so far, we need to further develop and test this area in the future.
	 *
	 * @since   1.0.0
	 *
	 * @see WPDA::set_option()
	 * @see WPDA_Repository::create()
	 * @see WPDA_Upgrade
	 */
	protected static function activate_blog() {

		if ( current_user_can( 'activate_plugins' ) ) {

			// Check if plugin has already been active (first array element).
			$wpda_version_installed = get_option( WPDA::OPTION_WPDA_VERSION[0] ); // Read directly from wp_options.

			if ( ! $wpda_version_installed ) {
				// Create plugin repository.
				$wpda_repository = new WPDA_Repository();
				$wpda_repository->create();
				// Save plugin version.
				WPDA::set_option( WPDA::OPTION_WPDA_VERSION );
			} else {
				// Read from WPDA class to compare with version in wp_options.
				$wpda_version_current = WPDA::get_option_default( WPDA::OPTION_WPDA_VERSION );

				if ( $wpda_version_installed !== $wpda_version_current ) {
					// Looks like there was a newer/older version installed.
					// Check if there is an upgrade/downgrade function in WPDA_Repository.
					$function_upgrade_name =
						'upgrade__' . str_replace('.', '_', $wpda_version_installed ) .
						'__' .str_replace('.', '_',  $wpda_version_current );
					if ( method_exists( '\WPDataAccess\Utilities\WPDA_Upgrade', $function_upgrade_name ) ) {
						// Specific upgrade function should perform the necessary steps.
						WPDA_Upgrade::$function_upgrade_name();
						// Save plugin version.
						WPDA::set_option( WPDA::OPTION_WPDA_VERSION );
					} else {
						// Just create the repository!
						// Create repository scripts are allways written in such a way that they'll try to save data.
						// Only in cases where this is not possible an upgrade will be performed.
						$wpda_repository = new WPDA_Repository();
						$wpda_repository->create();
						// Save plugin version.
						WPDA::set_option( WPDA::OPTION_WPDA_VERSION );
					}
				}
			}
		} else {

			wp_die( esc_html__( 'ERROR: Not authorized', 'wp-data-access' ) );

		}

	}

	/**
	 * Deactivate plugin WP Data Access
	 *
	 * On deactivation we leave the repository and options as they are in case the user wants to reactivate the
	 * plugin later again. Tables and options are deleted when the plugin is uninstalled. To keep tables and options
	 * on uninstall change plugin settings (see uninstall settings).
	 *
	 * @since   1.0.0
	 */
	public static function deactivate() {

		if ( ! current_user_can( 'activate_plugins' ) ) {

			wp_die( esc_html__( 'ERROR: Not authorized', 'wp-data-access' ) );

		}

	}

}
